# Stubs for flask.logging (Python 3.6)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from .globals import request
from typing import Any

def wsgi_errors_stream(): ...
def has_level_handler(logger: Any): ...

default_handler: Any

def create_logger(app: Any): ...
